c read/ save variables in static space -- to a LRU cache
c it is recommended that one thread handle writing and pruning,
c though many threads can read

c note: !PRUNE, !PRUNE2, !RESET and addition of entries should
c be serialized, so that only 1 can be active.
c Lookups need NOT be serialized

c$define INCL_REXXSAA
c$include rexxsaa.fap

! Declare our exported function.  Export it!

c$pragma aux (RexxFunctionHandler) SRE_MD5S_CACHE "SRE_MD5S_CACHE"

! SRE_MD5S_CACHE -- store variables 

c$noreference
	integer function SRE_MD5S_CACHE( name, numargs, args,
     &                            queuename, retstr )
c$reference

	character*(*) name, queuename
	integer numargs
	record /RXSTRING/ retstr, args(numargs)
        
	include 'rxsutils.fi'
        include 'fsublib.fi'

	character*(*) tmp,arg1,arg2,arg3

        character *2 crlf

        character *60 varnames(:),aname
        character *60 values(:),value1

        integer *2 hash_index(2048)             ! the hash table
        integer *2 index_list(:)
        integer timstamp(:),hitct(:)

        character *250 goober
        save varnames,values,hash_index,index_list,timstamp
        volatile varnames,values,hash_index,index_list,timstamp

        integer maxglobals,nextfree
        save maxglobals,nextfree
        volatile maxglobals,nextfree

        data maxglobals,nextfree/-1,-1/

        logical qmatch
        integer setlength,mm,tmplen,ierr1
        integer itmp,ido,idowas
        integer nn,critkilled,crittime,nkilled,istamp
        integer newest,oldest,iat,iat2,iat0,iattmp

        integer hashval,fig_hashval,time_stamp,remove_mode,writelock
        save writelock
        volatile writelock
        data writelock/0/

        integer mm3,mm2                                

        tmplen=190
        deallocate(tmp,stat=ierr1)
        allocate(tmp*tmplen)

c gotta be a few arguments
	if( numargs .lt. 1.or.numargs.gt.3 )then
           SRE_MD5S_CACHE = 443
           return
	endif

c are we "locked"?
        if (writelock.eq.1) then    !  wait for lock to be released
           tmp=' '                      ! not released, so give up
           do mm=1,50           ! wait for up to 5 seconds
             call wait_asec(0.1)    
             if (writelock.ne.1)  exit !lock has been released?
           enddo
           if (writelock.eq.1) then  ! still locked? 
              tmp=' '               
              goto 2000
           endif
        endif


        allocate( arg1*args(1).strlength, location=args(1).strptr )
        call strip_arg(arg1)

        setlength=0
        if (numargs.gt.1) then
         if (args(2).strlength.gt.0) then          !get the value 
            allocate( arg2*args(2).strlength, location=args(2).strptr )
            setlength=args(2).strlength
         else
           allocate(arg2*1)
           arg2=' '
         endif
       endif

       if (numargs.eq.3) then
          if (args(3).strlength.gt.0) then          !get the value 
              allocate( arg3*args(3).strlength, location=args(3).strptr )
           else
              allocate(arg3*1)
              arg3=' '
           endif
        endif

c Initialize? 
        if (arg1.eq.'!STARTUP' .or. arg1.eq.'!RESET') then
           writelock=1

           if (arg1.eq.'!STARTUP') then  !set some cache size parameters 
             if (setlength.eq.0) then 
                maxglobals=1000                 ! default size of cache
                ierr1=0
             else
               read(arg2,*,iostat=ierr1)maxglobals
             endif
             if (ierr1.ne.0) then
               tmp='error_bad_md5maxcachesize'
               goto 2000
             endif
             IF (MAXGLOBALS.Gt.32000) THEN
               tmp='error_md5maxcachesize_mustbe_le_32000'
               goto 2000
             endif
             if (maxglobals.le.10) then
               tmp='error_md5maxcachesize_mustbe_gt_10'
               goto 2000
             endif
           endif
           if (arg1.eq.'!RESET') then
              call wait_asec(0.5)               ! pause for 1.5 seconds
           endif                                !so as to let others finish

c now do reseting (or initial setting )
           deallocate(varnames,stat=ierr1)
           allocate(varnames(maxglobals),stat=ierr1)
           if (ierr1.ne.0) then
              tmp='error_allocating_md5cache_varnames'
              goto 2000
           endif

           deallocate(values,stat=ierr1)
           allocate(values(maxglobals),stat=ierr1)
           if (ierr1.ne.0)  then
              tmp='error_allocating_md5cache_values'
              goto 2000
           endif

           deallocate(INDEX_LIST,stat=ierr1)
           allocate(index_list(maxglobals),stat=ierr1)
           if (ierr1.ne.0) then 
               tmp='error_allocating_md5cache_indexlist'
               goto 2000
           endif

           deallocate(timstamp,stat=ierr1)
           allocate(timstamp(maxglobals),stat=ierr1)

           if (ierr1.ne.0) then 
               tmp='error_allocating_md5cache_timstamp'
               goto 2000
           endif

           deallocate(hitct,stat=ierr1)
           allocate(hitct(maxglobals),stat=ierr1)

           if (ierr1.ne.0) then 
               tmp='error_allocating_md5cache_hitct'
               goto 2000
           endif

c now initialize index_list (to point to next free
           do mm=1,maxglobals-1
              index_list(mm)=mm+1
              timstamp(mm)=0
              hitct(mm)=0
           enddo
           index_list(maxglobals)=0                      !0 signals "none free "
           timstamp(maxglobals)=0                      
           hitct(maxglobals)=0                      

           do mm=1,2048
              hash_index(mm)=0
           enddo
          
           nextfree=1                           ! 1st element is free

           tmp='ok'
           goto 2000            !ok
        endif


c info?
        if (arg1.eq.'!STATUS') then 
           execute do_info1
           goto 2000
        endif

        if (arg1.eq.'!INFO') then 
           execute do_info2
           goto 2000
        endif

        if (arg1.eq.'!PRUNE'.or.arg1.eq.'!PRUNE2') then 
           writelock=1
           call wait_asec(0.5)               ! pause for 1.5 seconds
           execute do_prune

           goto 2000
        endif
       
        if (arg1.eq.'!REMOVE'.and.numargs.gt.1) then
             deallocate(arg1,stat=ierr1)
             iat0=len(arg2)
             allocate( arg1*iat0)
             arg1=arg2
             remove_mode=1
        endif

c  Note: global variables are not robust against rapid changes by many threads

        if (nextfree.eq.-1) then                       !was cache initialized? 
            tmp='error_md5cache_not_initialized'
            goto 2000
        endif

c 1) compute hash value (twixt 1 and 2048)
     
       hashval=fig_hashval(arg1)

          
c 2) see if an entry exists
       if (hash_index(hashval).eq.0) then     !no such hash, hence no such variable
          if (remove_mode.eq.1 .or. numargs.eq.1) then   !lremove or lookup -- so return ' '
            tmp=' '
            if (remove_mode.eq.1) then
                write(tmp,871)arg1,hashval
 871            format('NO_SUCH_ENTRY:',a,i6)
            endif                
            goto 2000
          endif                               !otherwise, add a variable & value 
          if (nextfree.eq.0) then
              tmp=' '                         ! ' ' signals "no more room 
              if (numargs.eq.3)tmp=arg3
              goto 2000
          endif


c if here, no such hash, not remove mode, and room to expand. So add hash table entry, and entry for this stuff
          hash_index(hashval)=nextfree          ! get address of next free slot

          itmp=index_list(nextfree)             ! this will be the new next free slot
          index_list(nextfree)=0                ! signal that this is end of "hash-valued" list
          varnames(nextfree)=arg1               ! save the variable
          values(nextfree)=arg2                 ! and it's value        
          timstamp(nextfree)=time_stamp()
          hitct(nextfree)=1
          nextfree=itmp                         ! Finally, set nextfree
          tmp=arg2
          goto 2000
        endif

c got a "hash-valued" list -- check each one
        qmatch=.false.
        ido=hash_index(hashval)

        do while (ido.ne.0)                      !0 signals "end of list 

           idowas=ido

           if (timstamp(ido).eq.0.and.numargs.eq.1) then   !timstamp of 0 means "voided entry"
                ido=index_list(ido)                        !which will be deleted on next prune
                cycle
           endif

           aname=varnames(ido)

           if (aname(1:3).eq.arg1(1:3)) then     ! preliminary check
              if (aname.eq.arg1) qmatch=.true.
           endif
           if (.not.qmatch) then
                ido=index_list(ido)                   
                cycle                   !MIGHT EXIT, IF IDO=0 (END OF LIST) OCCURS
           endif
 
c if here, found a match!
          if (remove_mode.eq.1) then       ! avoid rare problems
              timstamp(ido)=0                   ! mark as void, and to be pruned       
              tmp=values(ido)
              goto 2000
           endif

           if (numargs.eq.1) then                ! lookup 
                tmp=values(ido)
                timstamp(ido)=time_stamp()      !update lru
                hitct(ido)=hitct(ido)+1
                goto 2000
           else                                  ! or set?
                values(ido)=arg2
                timstamp(ido)=time_stamp()      !update lru
                hitct(ido)=hitct(ido)+1
                tmp=arg2
                goto 2000
           endif

        enddo                  ! scanning this  list at hashval

c if here, no match in non-empty hash list
        if (numargs.eq.1) then                          !lookup? return ' '
             tmp=' '
             goto 2000
        endif
       if (nextfree.eq.0) then   ! out of space -- 
            tmp=' '                 !return "out of space" flag 
            if (numargs.eq.3)tmp=arg3   !if specified
            goto 2000
       endif


       index_list(idowas)=nextfree  
       itmp=index_list(nextfree)
       index_list(nextfree)=0
       varnames(nextfree)=arg1
       timstamp(nextfree)=time_stamp()
       hitct(nextfree)=1
       values(nextfree)=arg2
       nextfree=itmp
       tmp=arg2

       goto 2000

c jump here to return status code stored in tmp
 2000  tmplen=max(1,lentrim(tmp))
       call CopyResult( tmp,tmplen, retstr )
       deallocate( tmp )
       deallocate(arg1,stat=ierr1)
       deallocate(arg2,stat=ierr1)
       SRE_MD5S_CACHE = VALID_ROUTINE
       if (writelock.eq.1) writelock=0              ! done with update?
       return



c--------------------
c prune -- zap old entries 

        remote block do_prune
c 1) find oldest and newest entry
        newest=0 
        oldest=999999999
        crittime=0                              !removes !REMOVED entries
        critkilled=maxglobals                   !no limite on removals

        if (arg1.ne.'!PRUNE2') then             !this is a generic PRUNE
          do nn=1,maxglobals
             if (timstamp(nn).eq.0) cycle        
             if (timstamp(nn).gt.newest) newest=timstamp(nn)
             if (timstamp(nn).lt.oldest) oldest=timstamp(nn)
          enddo
          crittime=oldest+((newest-oldest)/3)
          critkilled=maxglobals/3
        endif

c 2) remove oldest "1/3" -- that is, entries that are older then the 1/3 point
c     between oldest and newest 

        nkilled=0                       !or, stop when maxglobals/3 entries have been removed

        do nn=1,2048 
             iat=hash_index(nn)

             do while (iat.ne.0)                    ! might have to recursively clear hash indexes
               istamp=timstamp(iat)
               if (istamp.le.crittime)  then       ! clear the "hashindex
                   timstamp(iat)=0                   ! "clear" entry
                   hitct(iat)=0
                   hash_index(nn)=index_list(iat)    ! 2nd element becomes top of list (possibly 0)
                   index_list(iat)=nextfree       !this is top of next free list
                   nextfree=iat                     
                   nkilled=nkilled+1
                   if (nkilled.gt.critkilled) goto 200     ! enough have been removed 
                   iat=hash_index(nn)                    ! the location of the new head of hash
                else                        ! hashindex is okay */
                   exit                       ! now check rest of list
                endif
             enddo

c now do none hash-index elements of list 

             iat0=iat                        ! the 'prior' element (starting with the first 
             iat=index_list(iat)             ! the next element

             do while (iat.ne.0)
                if (iat.eq.0) cycle                !nothing more in this list         
                istamp=timstamp(iat)
                if (istamp.le.crittime)  then      ! clear the "hashindex
                   timstamp(iat)=0                 ! "clear" entry
                   hitct(iat)=0
                   iattmp=iat
                   iat=index_list(iat)               ! the next element
                   index_list(iat0)=iat              ! prior element points to next element

                   index_list(iattmp)=nextfree       !this is now the top of next-free list
                   nextfree=iattmp                   !so point nextfree list to here      
                   nkilled=nkilled+1
                   if (nkilled.gt.critkilled) goto 200     ! enough have been removed 
                else
                   iat0=iat                       ! this is now the "prior element 
                   iat=index_list(iat)            ! the next element
                endif
            enddo

        enddo                   !check all 2048 hash elements

 200    continue
        write(tmp,98)nkilled
 98     format(i10)

        exit

        end block


c------------- Info  ----
       remote block do_info1 

        crlf(1:1)=char(13)
        crlf(2:2)=char(10)

        itmp=0
        do mm=1, maxglobals
          if (timstamp(mm).gt.0) itmp=itmp+1
        enddo
        write(tmp,983)maxglobals,itmp
 983    format(' Max & current # of cache entries=',i8,', ',i8)
        exit
        end block



c------------- Info  ----
c return Timestamp #hits Varname Value crlf ....
c read as parse var timestamp numhits varname value '0d0a'x ...
c (assuming no crlfs in the value)

       remote block do_info2  

        crlf(1:1)=char(13)
        crlf(2:2)=char(10)

        tmplen=max(80,maxglobals*400)
        deallocate(tmp,stat=ierr1)
        allocate(tmp*tmplen,stat=ierr1)
        if (ierr1.ne.0) then
           tmp='error_allocate_md5cache'
            exit
        endif
        itmp=0
        do mm=1, maxglobals
          if (timstamp(mm).gt.0) itmp=itmp+1
        enddo
        write(tmp,9831)itmp,crlf
 9831    format(i8,a)
        iat=11       
        do mm=1, maxglobals
          if (timstamp(mm).gt.0) then
             value1=values(mm)
             do mm2=1,lentrim(value1)
                 mm3=ichar(value1(mm2:mm2))
                 if (mm3.eq.10.or.mm3.eq.13) value1(mm2:mm2)=' '
             enddo
             write(goober,89)timstamp(mm),hitct(mm),varnames(mm),
     1                  value1,crlf
 89          format(i10,1x,i10,1x,a,1x,a,a)        
             iat2=lentrim(goober)+2+iat
             tmp(iat:iat2)=goober
             iat=iat2+1
          endif
        enddo
        exit
        end block

        end

  

