c read/ save variables in static space (system vars)

c$define INCL_REXXSAA
c$include rexxsaa.fap

! Declare our exported function.  Export it!

c$pragma aux (RexxFunctionHandler) SRE_SYSVARS "SRE_SYSVARS"

! SRE_SYSVARS -- store variables 

c$noreference
	integer function SRE_SYSVARS( name, numargs, args,
     &                            queuename, retstr )
c$reference

	character*(*) name, queuename
	integer numargs
	record /RXSTRING/ retstr, args(numargs)
        
	include 'rxsutils.fi'

	character*(*) tmp,arg1,arg2
        CHARACTER *2 CRLF        
        character *232 aline

        character *25  globnames(:)
        integer *2 globindx(:)
        integer *2 hashtable(2048)

        character *132 globvals(:)
        integer nglobals
        save globnames,globvals,nglobals  !for storing global variables
        save hashtable,globindx
        volatile globnames,globvals,nglobals  !for storing global variables
        volatile hashtable,globindx

        integer maxglobals
        save maxglobals
        volatile maxglobals

        data maxglobals/-1/

        integer setlength,mm,tmplen,ierr1,II

        integer jash,indx1,find_hashtable,fig_hashval

        tmplen=132
        deallocate(tmp,stat=ierr1)
        allocate(tmp*tmplen)

c gotta be a few arguments
	if( numargs .lt. 1.or.numargs.gt.3 )then
           SRE_SYSVARS = 443
           return
	endif

        allocate( arg1*args(1).strlength, location=args(1).strptr )
        call strip_arg(arg1)

c        write(6,*)'in globals ',arg1

        setlength=0
        if (numargs.gt.1) then
         if (args(2).strlength.gt.0) then          !get the value 
            allocate( arg2*args(2).strlength, location=args(2).strptr )
            setlength=args(2).strlength
         else
            allocate(arg2*1)
            arg2=' '
         endif
       endif

        if (arg1.eq.'!STARTUP') then
           nglobals=0
           if (setlength.eq.0) then 
                maxglobals=100
                ierr1=0
           else
               read(arg2,*,iostat=ierr1)maxglobals
           endif
           if (maxglobals.gt.32000) then   !32k is largest Integer *2 value
                tmp='error_maxsysvars_gt_32000'
                goto 2000
           endif

           tmp='ok'
           if (ierr1.ne.0) tmp='error_reading_maxsysvars'
           deallocate(globnames,stat=ierr1)
           deallocate(globvals,stat=ierr1)
           deallocate(globindx,stat=ierr1)

           allocate(globindx(maxglobals),stat=ierr1)
           if (ierr1.ne.0) then 
                tmp='error_allocating_globindx'
                goto 2000
           endif

           allocate(globnames(maxglobals),stat=ierr1)
           if (ierr1.ne.0) then 
                tmp='error_allocating_globnames'
                goto 2000
           endif

           allocate(globvals(maxglobals),stat=ierr1)
           if (ierr1.ne.0) then 
                tmp='error_allocating_globvals'
                goto 2000
            endif

           if (tmp.eq.'ok') then 
             do mm=1,2048
                hashtable(mm)=0
             enddo
             do mm=1,maxglobals
                globindx(mm)=0
             enddo
           endif

           goto 2000
        endif

        if (maxglobals.lt.0) then
          tmp='error_sysvars_not_initialized'
          goto 2000
        endif

! INFO :: info on what's stored

        if (arg1.eq.'!INFO') then 
           execute do_info
           goto 2000
        endif
           

c  Note: global variables are not robust against rapid changes by many threads
c        write(6,*)'checking ',nglobals,maxglobals


c compute hash (1-2048)
        jash=fig_hashval(arg1)

        indx1=find_hashtable(jash,arg1,hashtable,globnames,
     1                        globindx,maxglobals)

C MATCH not found 
        if (indx1.eq.0) then            ! match not found

          if (setlength.eq.0) then              !not setting 
             if (numargs.gt.2) then           ! use a "n.a. flag"
                deallocate(tmp,stat=ierr1) 
                allocate(tmp*args(3).strlength, location=args(3).strptr )
             else
                tmp=' '                    !return ' ' (if not setting )
             endif
             goto 2000
          
          else                 !adding new var

            if (nglobals.ge.maxglobals) then
                tmp='error_too_many_sysvars'
                goto 2000
            endif
            write(tmp,*,iostat=ierr1)arg2
            if (ierr1.ne.0) then                         ! error writing value
                tmp='error_writing_sysvar_1st'
                goto 2000
            else                                 !okay
               nglobals=nglobals+1
               globvals(nglobals)=arg2
               globnames(nglobals)=arg1
               globindx(nglobals)=hashtable(jash)  !may be 0, may be current list-with-this-hash
               hashtable(jash)=nglobals             !head of list becomes current variable
            endif
          endif

C MATCH was found 
        else                             ! match WAS found 
            
           if (setlength.gt.0) then             !set new value
               globvals(indx1)=arg2
            endif
            write(tmp,*,iostat=ierr1)globvals(indx1)
            if (ierr1.ne.0) then                         ! error writing value
                tmp='error_writing_sysvarvalue'
                goto 2000
            endif

        endif

c jump here to return status code stored in tmp
 2000  tmplen=max(1,lentrim(tmp))
        call CopyResult( tmp,tmplen, retstr )
       deallocate( tmp )
       deallocate(arg1,stat=ierr1)
       deallocate(arg2,stat=ierr1)
       SRE_SYSVARS = VALID_ROUTINE
       return



c------------- Info  ----
       remote block do_info  

        deallocate(tmp,stat=ierr1)

        II=180+(250*NGLOBALS)
        ALLOCATE(TMP*II)

        crlf(1:1)=char(13)
        crlf(2:2)=char(10)

        write(tmp,991,iostat=ierr1)nglobals,maxglobals,
     1         crlf,crlf,crlf
 991    format('  Using ',i6,
     1         ' of ',i6,' possible variables ',a,
     1      'Variable                     :     Value ',a,
     1      '---------------------------------------------',a)

        if (ierr1.ne.0) then
           tmp='error_writing_status_info'
           exit
        endif

        do mm=1,nglobals
           write(aline,992)globnames(mm),
     1           globvals(mm)(1:lentrim(globvals(mm))),crlf
 992       format(2x,a25,2x,': ',a,a)
           CALL ADD_STRING(TMP,ALINE)
        enddo
        exit
        end block
        end


c--------------
c capitalize and strip
        subroutine strip_arg(sent)
        character *(*) sent
        integer ij,i1,let
        i1=0
        do  ij=1,lentrim(sent)
            let=ichar(sent(ij:IJ))
            if (let.gt.32.and.i1.eq.0) i1=ij
            if (let.ge.97.and.let.le.122) sent(ij:ij)=char(let-32)
        ENDDO
        if (I1.EQ.0) then
          sent=' '
        else
          sent=sent(i1:lentrim(sent))
        endif
        return
        end



c------------------
c compute hash
        integer function fig_hashval(astring)
        character *(*) astring
        integer isum,mm
        isum=0
        do mm=1,lentrim(astring)
           isum=isum+ichar(astring(mm:mm))
        enddo
        fig_hashval=mod(isum,2048)+1
        return
        end

c-----------------------------
c find arg1 in list starting from hashtable(jash)
        integer function find_hashtable(jash,arg1,hashtable,
     1                             globnames,globindx,maxglobals)
        integer jash,maxglobals,mm
        character *(*) arg1
        integer *2 hashtable(2048),globindx(maxglobals)
        character *25 globnames(maxglobals)
        integer ilook
        
        find_hashtable=0                !default is "not found"
        if (jash.eq.0) return

c look thorough list that starts at hashtable(jash)
        ilook=hashtable(jash)
        if (arg1.eq.globnames(ilook)) then
            find_hashtable=ilook
            return
         endif

c not at head of list, try latter ones
        do mm=1,5000000           !do forever
          ilook=globindx(ilook)
          if (ilook.eq.0) return                ! not found 
          if (arg1.eq.globnames(ilook)) then
              find_hashtable=ilook
              return
           endif
        enddo

        return          !should NEVER get here 
        end  

