/* Sre2000, ver 1.01 

This is the main program. It reads the initialization file, 
launches a "status" daemon, stores the "parent" and "child" procedures
to macrospace (for use as transient request-specific daemons), loads the 
main "filter" to macrospace, and launches filter-specific startup procedure and
daemon manager.

*/

/***** User configurable parameters -- they usually do not need to be changed ***/

/* The SRE2000_CFG file contains several basic (not filter specific) parameters */
SRE2000_CFG='SRE2000.CFG'

/* if TZ variable not available, or not understood, use this
   GMT offset (in seconds). Note that this is the number of
   seconds that should be subtracted from local time to obtain GMT time.
   Example: -18000 for EST */
default_gmtoffset=-18000

/* Frequency (in seconds) to update status information */
update_rate=1

/* "target" time in seconds-- (if average response time>target_time, change color in display */
target_time=1

/* show the current time (every update_rate seconds)*/
clockon=0

/* beep when a connection is opened */
beepon=0


/*****  End of User configurable parameters ****/


/* 1) Load the various dlls, initializse, etc */

dmn_tid=0

parse arg aopts; aopts=strip(aopts)
call sre_ini aopts

signal on syntax name abend
signal on error name abend
signal on failure name abend
signal on halt name doquit

blank = ' '
null = ''
default_gmtoffset0=default_gmtoffset

val = SockInit()
if val \= 0 then do
   say "ERROR: SockInit failure: " errno
   exit
end
/* create a socket */
sock1 = SockSocket('AF_INET','SOCK_STREAM','IPPROTO_TCP')
if sock1 < 0 then do
   say 'ERROR: can not get socket :' errno
   exit
end

/* bind the socket to port */
addr.port =sreparams.!PORT 
addr.addr = INADDR_ANY
addr.family = AF_INET
prob=1
foo=socksetsockopt(sock1,'SOL_SOCKET','SO_REUSEADDR',prob)
val = SockBind(sock1,'addr.')
if val \= 0 then do
       say 'ERROR: Can not bind socket :' errno
       val = SockShutdown(sock1,2)
       val = SockClose(sock1)
       exit 
end

/* make this socket 'passive' */
inq=max(sreparams.!LIMITCLIENTS,8)
val = SockListen(sock1,inq)
if val \= 0 then do
  say 'ERROR: can not listen on socket :' errno
  val = SockClose(sock1)
  exit
end

if Sockioctl(sock1,'FIONBIO',1) = -1 then  do      /* switch to nonblocking mode */
   foo=sre_audit('SRE2000','crashed in sockioctl of sock1 'socket':'errno)
   signal errordone
end

/* ready to start accepting requsts. So invoke "status monitor" */

/*  invoke audit daemon */
foo=rexxthread('f','SRE2000.RX5',sreparams.!tracking_on, ,
              sreparams.!auditfile,auditqueue)
if foo=0 then do
   say "ERROR: Unable to launch audit daemon"
   exit
end 
say  ' Audit Daemon: started on thread 'foo

/* and status / cache controller */
foo=rexxthread('f','SRE2000.RX4', ,
              sreparams.!filter' 'sreparams.!fast_filter,sreparams.!port,pid,sreparams.!status_items,sock1,update_rate,target_time, ,
              thesem,SRE2000_CFG,clockon, ,
              cachequeue,sreparams.!req_cache_record,sreparams.!req_cache_hitfile, ,
              sreparams.!no_modify_onthefly,sreparams.!status_message,sreparams.!server_software)
if foo=0 then do
   say "ERROR: Unable to launch status-monitoring daemon"
   exit
end 
say  ' Status Daemon: started on thread 'foo


call syssleep(1.3)              /* wait a second */

schecked=0
transaction=0
show_wait=wordpos('SHOW_WAIT',sreparams.!status_items)

Do Forever              /**** Now enter infinite "accept requests" loop ***/

   /* wait for a connection on this socket */

   schecked=schecked+1

   is_update=eventsem_query(thesem_reset)  /* check reset semaphore */

   if is_update>0 |  schecked>200 then do     /* every 20 connections (with no pause), check status */
      call check_status is_update
      if result=1 then signal doquit
      schecked=0
   end 

   if show_wait>0 then do
      ttime=time("n")
      call charout, '1b5b'x||'1;2H'||cy_ye||'waiting: '||ttime||normal
   end

  kct=0
  do forever                    /* non blocking socket -- so we have to poll it */
     sock2 = SockAccept(sock1)
     if sock2>0 then leave              /* got a request */
     kct=kct+1
     if (kct//250)=1 then do     /* check for flags every 2.5 seconds */
         kct=2
         call check_status 
         if result=1 then signal doquit
         schecked=0
     end 
     call syssleep(0.01)                /* take a short break */
  end
 
   if sock2 < 0 then do
      say  "ERROR: Can not accept on socket :" errno
      val = SockClose(sock1)
      signal doquit
   end

   if beepon then call charout,'07'x

   
   if check_bad=1 then do
      gotbad=0
      cc = SockGetPeerName(sock2,"caddr.!")
      is1=space(caddr.!addr,0)

      if badips.is1=1 then do
         gotbad=1
      end
      else do
        if badips.!wild>0 then do
           parse var is1 p1 '.' p2 '.' p3 '.' p4
           t1=p1||'.*'
           gotbad=badips.t1
           if gotbad<>1 then do
                t1=p1||'.'||p2||'.*'
                gotbad=badips.t1
                if gotbad<>1 then do
                   t1=p1||'.'||p2||'.'||p3||'.*'
                   gotbat=badips.t1
                end
            end
        end                     /* checking wilds */
      end                       /* exact match is bad */
      if gotbad=1 then do
         rc=sockshutdown(sock2,2)
         rc = SockClose(sock2)
         if verbose>0 then call write_message('Denying connection from 'is1)
         iterate
      end
   end

/* check for # of active clients (actually, client handling "parent" threads */
/* determine "client number". If 0, then no room */

clinum=sre_localvars('!FREESLOT',sreparams.!limitclients)
if clinum=0 then do
   if wordpos('LIMITS',sreparams.!STATUS_ITEMS)>0 then do 
      call write_error "Client limit exceeded at "transaction 
      foo=sre_countvars('LIMITS')
   end 
   rc=sockshutdown(sock2,2)
   rc = SockClose(sock2)
   iterate
end

/* get numeric name of host */
   val = SockGetSockName(sock2,"saddr.!")
   if val \= 0 then do
      call write_error "ERROR: cannot get host name :" errno
      iterate
   end
/* get name of host */
   asaddr=saddr.!addr
   hname=hnames.asaddr
   if hname='' then do
      aa=sockgethostbyaddr(asaddr,"ahost.!")
      hname=ahost.!name
      if aa=0 then hname=asaddr
      hnames.asaddr=hname
   end

/* fire up thread to handle this request */
   transaction=transaction+1
   ccs=sreparams.!req_cache_enable','sreparams.!req_cache_verify','sreparams.!req_cache_call_anyways','sreparams.!req_cache_record
   lims=sreparams.!limitbody' 'sreparams.!limitheader' 'sreparams.!limittimewait' 'sreparams.!limittimeinactive' 'sreparams.!limittimetotal
   foo=rexxthread('m','SRE_MOD2',clinum,pid,sock2,asaddr,saddr.!port,hname, ,
                     sreparams.!default_socket_timeout,transaction, ,
                     sreparams.!port,lims,sreparams.!status_items, ,
                     sreparams.!audit_items,ccs,sreparams.!fast_filter)
end

exit


abend:
        ln = sigl
        erc = rc
        say 'Syntax error, line' ln':' errortext(erc)
        say '+++' sourceline(ln)
        exit

doquit:
  say normal
  say
  call charout, '1b5b'x||'19;1H'|| '1b5b'x||'K'
  call charout, '1b5b'x||'18;1H'|| '1b5b'x||'K'
  call charout, '1b5b'x||'17;1H'|| '1b5b'x||'K'
 say "Leaving Sre2000 "
 foo=sockshutdown(sock1,2)
 foo=sockclose(sock1)
 im=queued()
 do mm=1 to im
    pull .
 end 

/* kill all the threads */
/* how many threads are open... */
threadlist=RxListThreads()
nowtid=dostid()
nkilled=0
do mm=1 to words(threadlist)
   atid=strip(word(threadlist,mm))
   if nowtid=atid then iterate
   wow=rxprckillthread(atid)
   if wow<>0 then
      say "Unable to kill thread: "atid ' ('wow
    else
        nkilled=nkilled+1
end 
say "# of threads killed= "nkilled "( of "words(threadlist)-1")"
exit



/********/
/* status check */
check_status:
parse arg is_update

/* check for "pause */
do forever
   aa=sre_pause()
   if aa<>0 & aa<>'' then do   /* pause? wait 1/10 second, and check again */
       foo=syssleep(0.1)
       iterate
   end 
   leave
end

  aa=sre_shutdown()  /* check for shutdown signal */
  if aa<>'' & aa<>'0' then do 
     rc=sockshutdown(sock2,2)
     rc = SockClose(sock2)
     rc=sockshutdown(sock1,2)
     rc = SockClose(sock1)
     if translate(aa)='X' then do  /* wait for 90 seconds, or until all requests are done */
        do mm=1 to 90
           if sre_localvars('!CLIENTS')=0 then leave
           call syssleep(1)
        end
     end
     return 1
  end 


/* check for "reset" signal */
/* check for a "reset parameters" (as turned on by the daemon in 'SRE2000.rx4) */  

if is_update='' then
   is_update=eventsem_query(thesem_reset)  /* check reset semaphore */

if is_update>0 then do
   aa=eventsem_reset(thesem_reset)
   call update_sre_ini
   if result=0 then return 1            /* some kind of error */
   if pos('RESET',sreparams.!audit_items)>0 then foo=sre_audit('RESET ',time('n')||' '||date('n'),0,1)
end 

if sreparams.!Unallowed_file<>0 then do
  aa=stream(sreparams.!unallowed_file,'c','query datetime')
  if aa<>badips.!datetime then do
     call read_unalloweds
     if verbose>0 then call write_message '# of unallowed IP entries ='||badips.!total
  end 
end 

return 0

/******************************/
/* initialize, load various dlls, etc. */
sre_ini:

crlf='0d0a'x

basicqueue=rxqueue('G')

/* currently required variables in SRE2000_CFG file */
srevars='PORT DATADIR LIMITBODY LIMITCLIENTS LIMITHEADER LIMITTIMETOTAL ' ,
        'LIMITTIMEWAIT LIMITTIMEINACTIVE FILTER FAST_FILTER DAEMON_MANAGER INIT_PROC ' ,
        'DEFAULT_SOCKET_TIMEOUT  SERVER_SOFTWARE STATUS_ITEMS AUDITFILE AUDIT_ITEMS ',
        'SERVER_SOFTWARE_SHORT USER_STATS UNALLOWED_FILE NO_MODIFY_ONTHEFLY ' ,
        'MAXTHREADS MAXHOSTS MAXLOCALS MAXSYSVARS MAXMD5S MAXCOUNTERS MAXCLIENTNAMES ' ,
        'MAXSTATVALS REQ_CACHE_SIZE REQ_CACHE_CALL_ANYWAYS REQ_CACHE_ENABLE ',
        'REQ_CACHE_VERIFY REQ_CACHE_RECORD REQ_CACHE_HITFILE STATUS_MESSAGE TRACKING_ON '

/* enable ansi screen controls */
ansion=checkansi()
if ansion=1 then do
  aesc='1B'x
  cy_ye=aesc||'[37;46;m'
  cyanon=cy_ye
  normal=aesc||'[0;m'
  bold=aesc||'[1;m'
  re_wh=aesc||'[31;47;m'
  reverse=aesc||'[7;m'
  call charout, '1b5b'x||'2J'

end
else do
  say " Warning: Could not detect ANSI....  Install will look ugly ! "
  cy_ye="" ; normal="" ; bold="" ;re_wh="" ;
  reverse=""
end  

if aopts='?' then do
    call showopts
    exit
end 

say "            "cy_ye"Sre2000 ver 1.01 "normal
say "    To view options. enter "bold"x:>SRE2000 ? "normal


say
say bold" Initializing ... "normal

/* Load various dlls */
foo=rxfuncquery('sysloadfuncs')
if foo=1 then do
  foo=RxFuncAdd( 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs')
  if foo=1 then do
     say "ERROR: could not find RexxUtil.DLL " 
     exit
  end /* do */
  call SysLoadFuncs
   
end

parse upper version rexxtype .
os2ver=sysos2ver()
say "  Using REXX version: " rexxtype ', under OS/2 ver 'os2ver


foo=rxfuncquery('rexxlibregister')
if foo=1 then do
 foo=rxfuncadd( 'rexxlibregister','rexxlib', 'rexxlibregister')
  if foo=1 then do
     say "ERROR: could not find RexxLib.DLL " 
     exit
  end /* do */
 call rexxlibregister
end

foo=RxFuncQuery("SockLoadFuncs") 
if foo=1 then do
    foo=RxFuncAdd( "SockLoadFuncs","rxSock","SockLoadFuncs")
    if foo=1 then do
       say "ERROR: could not find rxSock.DLL " 
       exit
    end /* do */
    call SockLoadFuncs
end
ver=sockversion()
if ver<2.0 then do
   say 'WARNING: old version of rxsocket ('ver'). We recommend using 2.0 or above'
end 


if rxfuncquery('srx_md5')=1  then do
  foo=RXFuncAdd( 'SRXLoadFuncs', 'SRXFUNC', 'SRXLoadFuncs')
   if foo=1 then do
       say "ERROR: could not find srxFunc.DLL " 
       exit
   end /* do */
  call SRXLoadFuncs
end
if rxfuncquery('srx_md5')=1  then do
       say "ERROR: could not find srx_Md5 in srxfunc.DLL " 
       exit
end

/* Load the rxgzlib functions */
if rxfuncquery('rxgzLoadFuncs')=1 then do
   foo=RxFuncAdd( 'rxgzLoadFuncs', 'rxgzLIB', 'rxgzLoadFuncs')
   if foo=1 then do
       say "ERROR: could not find rxGZlib.DLL " 
       exit
   end 
   call rxgzLoadFuncs
end 

if rxfuncquery('RxPrcLoadFuncs')=1 then do
 foo=RxFuncAdd('RxPrcLoadFuncs', 'RXPROC', 'RxPrcLoadFuncs')
 if foo=1 then do
    say "ERROR: could not find rxProc.DLL "
   exit
 end 
 call RxPrcLoadFuncs
end


if rxfuncquery('SRE_REQ_CACHE')=1  then do
  FOO=RXFuncAdd('SRELoadFuncs', 'SRE2000', 'SRELoadFuncs')
  if foo=1 then do
       say "ERROR: could not find 'SRE2000.DLL " 
      exit
  end 
  call SRELoadFuncs
end

/* load basic .rxl file */
/* find list of procedures */

arf='SRE2000.IDX'
foo=stream(arf,'c','query size')
if foo=0 | foo=' ' then do
   say "ERROR: unable to clear procedures from macrospace"
   exit
end 
else do                 /* zap old versions of procedures */
   afoo=stream(arf,'c','open read')
   aa=charin(arf,1,foo)
   afoo=stream(arf,'c','close')

   idid=0
   do until aa=''
      parse var aa aline '0d0a'x aa
      if aline='' then iterate
      aline=strip(aline)
      if abbrev(aline,';')=1 then iterate
      if words(aline)>1 then iterate
      foo=macrodrop(aline)
      if foo<>1 then do
            if macroquery(aline)<>' ' then do
              say "ERROR: unable to macro_drop " aline
              exit
            end         /* otherwise, it's not there */
      end 
      else do
          idid=idid+1
      end
   end 
   say '  'idid' old versions of macrospace procedures removed.'
end 


/* zap prior procedures */

if macroquery('SRE_COMMAND')="" then do
  rxlfile='SRE2000.RXL'
  if stream(rxlfile,'c','query exists')='' then do
      say "ERROR: "|| rxlfile ||" not available"
      exit
  end 
  aa=macroload(rxlfile)
  if aa=0 then do
      say "ERROR: "|| rxlfile ||" was not loaded ("|| aa
      exit
  end
end

taopts=translate(aopts)

/* alternative SRE2000_CFG file? */
if pos('-I ',taopts)>0 then do
   parse var taopts  '-I' SRE2000_CFG
   SRE2000_CFG=strip(SRE2000_CFG)
   say "   Using initialization file: " SRE2000_CFG
end 

if pos('-CVT',taopts)>0 then do   /* convert from goseve to Sre2000 */
  call charout, '1b5b'x||'2J'
  oo=gos2sre(' ')
  say
  say " Conversion(s) completed. "
  exit
end


/* modify initialization file (call MOD_INI.CMD) */
if pos('-M',taopts)>0 then do
  if pos('-MALL',taopts)=0 then do
     parse var taopts . '-M' daopt .
    if daopt=''  then daopt=SRE2000_CFG
    foo=mod_ini(daopt' 2')
    say
    say " Modifications completed. Status: "bold||foo||normal
    exit
  end
  parse var taopts . '-MALL' daopt .
  if daopt=''  then daopt=SRE2000_CFG
  foo=mod_ini(daopt,asrevars)
  say
  say " Modifications completed. Status: "bold||foo||normal
  exit
end

/* read initialization parameters */
foo=strip(translate(stream(SRE2000_CFG,'c','open read')))
if abbrev(foo,'READY')=0 then do
   say reverse'ERROR:'normal' can not open initialization file: 'SRE2000_CFG
   exit
end 
isz=stream(SRE2000_CFG,'c','query size')
if (isz=0 | isz='') then do
   say reverse'ERROR:'normal' empty or missing initialization file: 'SRE2000_CFG
   exit
end 

stuff=charin(SRE2000_CFG,1,isz)
foo=stream(SRE2000_CFG,'c','close')

sreparams.=''

/* now parse the file, eventually to save to SRE_SYSVARS cache */

do until stuff=''
   parse var stuff aline (crlf) stuff
   if aline='' then iterate
   aline=strip(translate(aline,' ','0d0a09'x))
   if abbrev(aline,';')=1 then iterate
   parse var aline avar '=' aval ; 
   avar=translate(strip(avar)) ; aval=strip(aval)
   if wordpos(avar,srevars)=0 then do
      say "  "reverse"Warning:"normal"Unknown SRE2000_CFG variable: "aline
      iterate
   end
   cname='!'||avar
   sreparams.cname=aval
end 


/* quality checks */

sreparams.!status_items=translate(space(sreparams.!status_items,1))
sreparams.!audit_items=translate(space(sreparams.!audit_items,1))

if datatype(sreparams.!default_socket_timeout)<>'NUM' then do
   say 'ERROR Bad socket timeout value : 'sreparams.!default_socket_timeout
   return 0
end 




nerrs=0
do mm=1 to words(srevars)
   aw=strip(word(srevars,mm))
   aw2='!'||aw
   if sreparams.aw2='' then do
      say "ERROR: Unspecified SRE2000_CFG variable: "aw
      nerrs=nerrs+1
   end
end /* do */
if nerrs>0 then exit            /* missing ini variables */

/* initialize  localvars */
foo=sre_localvars('!STARTUP',sreparams.!MAXTHREADS,sreparams.!MAXLOCALS)
IF FOO<>'1' then DO
   SAY "ERROR initializing sre_localvars: "foo
   exit
end

if words(srevars)+10 > sreparams.!maxsysvars then do
   say 'ERROR to many system variables('||words(srevars)||','sreparams.!maxsysvars||')'
   exit
end 

foo=sre_sysvars('!STARTUP',sreparams.!maxsysvars)
IF abbrev(FOO,'error')=1 then DO
   SAY "ERROR initializing sre_sysvars: "foo
   exit
end

/**
foo=sre_globalvars('!STARTUP',sreparams.!maxglobals)
IF abbrev(FOO,'error')=1 then DO
   SAY "ERROR initializing sre_globalvars: "foo
   exit
end
**/

appname='$SYSVARS_'||sreparams.!port
foo=sysini('SRE2000.INI',appname,'DELETE:')   /* initialize 'SRE2000.INI */
if foo='ERROR:' then do
   say "Unable to clear 'appname'in 'SRE2000 INI file "
   exit
end 


foo=sre_countvars('!STARTUP',sreparams.!maxcounters)
IF abbrev(FOO,'error')=1 then DO
   SAY "ERROR initializing sre_countvars: "foo
   exit
end


foo=sre_hostinfo('!STARTUP',sreparams.!maxhosts)
IF abbrev(FOO,'error')=1 then DO
   SAY "ERROR initializing sre_hostinfo: "foo
   exit
end

foo=sre_cache_ipname('!STARTUP',sreparams.!maxclientnames)
IF abbrev(FOO,'error')=1 then DO
   SAY "ERROR initializing sre_cache1: "foo
   exit
end

foo=sre_md5s_cache('!STARTUP',sreparams.!maxmd5s)
IF abbrev(FOO,'error')=1 then DO
   SAY "ERROR initializing sre_md5s: "foo
   exit
end


foo=sre_savevals('!STARTUP',sreparams.!maxstatvals,'TOTAL_TIME','START_TRANSMIT_TIME','BYTES')
IF abbrev(FOO,'error')=1 then DO
   SAY "ERROR initializing sre_savevals: "foo
   exit
end

if datatype(sreparams.!req_cache_size)<>'NUM' then sreparams.!req_cache_size=0
if sreparams.!req_cache_size>0 then do
  foo=SRE_REQ_CACHE('!STARTUP',sreparams.!req_cache_size)
  IF FOO<>'ok' then DO
     SAY "ERROR initializing sre_req_cache: "foo
     exit
  end
end

if sreparams.!req_cache_size=0 | datatype(sreparams.!req_cache_enable)<>'NUM' then sreparams.!req_cache_enable=0
if datatype(sreparams.!req_cache_record)<>'NUM' then sreparams.!req_cache_record=0


/* save globals to savevars */
do mm=1 to words(srevars)
   aw=strip(word(srevars,mm))
   aw2='!'||aw
   foo=sre_sysvars(aw,sreparams.aw2)
   if abbrev(foo,'error')=1 then do
        say 'ERROR unable to save 'aw' ('foo
        exit
   end 
end 

/* determine gmtoffset using TZ? */
tz=value('TZ',,'os2environment')
if tz='' then do
      say 'ERROR TZ variable is not available'
          say "    Using default GMTOFFSET= " default_gmtoffset
      say bold"    You should add TZ to CONFIG.SYS, or SET TZ before running Sre2000"normal
      sreparams.!gmtoffset=default_gmtoffset
end
else do
   sreparams.!gmtoffset=fig_gmtoffset(tz,default_gmtoffset0)
   say "  Using GMT Offset = "sreparams.!gmtoffset
   default_gmtoffset0=sreparams.!gmtoffset
end
oo=sreparams.!GMTOFFSET
foo=sre_sysvars('GMTOFFSET',oo)
foo=sysini('SRE2000.INI',appname,'GMTOFFSET',oo)


maindir=directory()
maindir=translate(maindir,'\','/')
maindir=strip(maindir,,'\')
foo=sre_sysvars('SERVDIR',maindir)
foo=sysini('SRE2000.INI',appname,'SERVDIR',maindir)
maindrive=filespec('D',maindir)

/* cache hit file */
if pos(':',sreparams.!req_cache_hitfile)=0 then do
   if abbrev(sreparams.!req_cache_Hitfile,'\')=0 then do
       sreparams.!req_cache_hitfile=maindir'\'sreparams.!req_cache_hitfile
       arf=stream(sreparams.!req_cache_hitfile,'c','query exists')
       if arf='' then do
            aa=';Sre2000 cache hit file '
            foo=charout(sreparams.!req_cache_Hitfile,aa,1)
            foo2=stream(sreparams.!req_cache_hitfile,'c','close')
            if foo=0 then do
                 say 'Cache-hit file created: 'sreparams.!req_cache_hitfile
            end 
            else do
                 say 'ERROR: unable to use, or create, cache-hit file: 'sreparams.!req_cache_hitfile
                exit
            end 
       end 
       foo=sre_sysvars('REQ_CACHE_HITFILE',sreparams.!req_cache_hitfile)
       say '  Using hit-cache file: 'sreparams.!req_cache_hitfile
   end
end


/* check existence of datadir */
datadir=sreparams.!DATADIR
datadir=translate(datadir,'\','/')
datadir=strip(strip(datadir,'t','\'))
if pos(':',datadir)=0 then do
   if abbrev(datadir,'\')=1 then do
        datadir=maindrive||datadir
   end
   else do
      datadir=maindir||'\'||datadir
   end /* do */
end
dd=dosisdir(datadir)
if dd=0 then do
   say 'ERROR: no such datadir ('datadir
   exit
end
sreparams.!datadir=datadir
foo=sre_sysvars('DATADIR',datadir)


/* check for existence of filter */
filter=stream(maindir||'\'||sreparams.!FILTER,'c','query exists')
if filter='' then do
   say 'ERROR: no such filter file ('filter
   exit
end


/* check for existence of filter */
fast_filter=sreparams.!FAST_FILTER
if FAST_FILTER<>0 then do
  fast_filter=stream(maindir||'\'||FAST_FILTER,'c','query exists')
   if fast_filter='' then do
     say 'ERROR: no such fast filter file ('fast_filter
     exit
   end
end
if fast_filter<>0 then do
  foo=macrodrop('SRE_FAST_FILTER')
  foo=macroadd('SRE_FAST_FILTER',fast_filter,'B')
  if foo<>1 then do
     say 'ERROR: unable to load fast filter ('fast_filter
     exit
  end
end


/* check limitbody and limitheader */
if datatype(sreparams.!limitbody)<>'NUM' then do
   say 'ERROR: LIMITBODY must be an integer > 0 '
   exit
end 
sreparams.!limitbody=sreparams.!limitbody*1000

/* check limitbody and limitheader */
if datatype(sreparams.!limitheader)<>'NUM' then do
   say 'ERROR: LIMITHEADER must be an integer > 0 '
   exit
end 
sreparams.!limitheader=sreparams.!limitheader*1000


say "  Using filter: "filter
if fast_filter<>0 then
   say '    and fast filter:'fast_filter||normal

say "  Using data directory: "datadir

/* set some counters to 0 */
alist='BYTESREADTOTAL  BYTESSENTTOTAL TRANSACTIONS  REQUESTS LIMITS ERRORS ' 
do mm=1 to words(alist)
   aww=strip(word(alist,mm))
   foo=sre_countvars(aww,0)
   IF abbrev(FOO,'error')=1 then do
      say 'ERROR initializing counter variable 'aww '='foo
      exit
   end 
end 

/* load sre_filter */
foo=macrodrop('SRE_FILTER')
foo=macroadd('SRE_FILTER',filter,'b')
if foo=0 then do
 say "ERROR could not load filter into macrospace "
 exit
end
/* load sre2000.rx3 */
foo=macrodrop("SRE_MOD3")
foo=macroadd("SRE_MOD3",'SRE2000.RX3','B')
if foo=0 then do
  say "ERROR could not load SRE2000.RX3 into macrospace "
  exit
end

/* load 'SRE2000.rx2 */
foo=macrodrop("SRE_MOD2")
foo=macroadd("SRE_MOD2",'SRE2000.RX2','B')
if foo=0 then do
  say "ERROR could not load 'SRE2000.RX2 into macrospace "
  exit
end


/* set startup date */
Foo=date('S')
adate=left(foo,4)||'.'||substr(foo,5,2)||'.'||substr(foo,7,2)
asec=time('N')
dstamp=adate||' '||asec
foo=sre_sysvars('LASTSTART',dstamp)
foo=sre_sysvars('LASTRESET',dstamp)
foo=sre_sysvars('LASTIDLE',dstamp)

pid=dospid()

hnames.=''

enmadd="SRE_"||sreparams.!port||"_"
foo=sre_sysvars('ENMADD',ENMADD)

thesem='\SEM32\'||enmadd
aa=eventsem_close(thesem)
aa=eventsem_create(thesem)
if aa<>0 & aa<>285 then do
 say "ERROR: could not create semaphore: "||thesem '('aa
 exit
end
aa=eventsem_reset(thesem)
if aa<0 & aa<>-300 then do 
 say "ERROR: could not create reset: "||thesem '('aa
 exit
end

thesem_reset=thesem||'RESET'
aa=eventsem_close(thesem_reset)
aa=eventsem_create(thesem_reset)
if aa<>0 & aa<>285 then do
 say "ERROR: could not create semaphore: "||thesem_reset '('aa
 exit
end
aa=eventsem_reset(thesem_reset)
if aa<0 & aa<>-300 then do 
 say "ERROR: could not create reset semaphore: "||thesem_reset '('aa
 exit
end

thesem_modify=thesem||'MODIFY'
aa=eventsem_close(thesem_MODIFY)
aa=eventsem_create(thesem_MODIFY)
if aa<>0 & aa<>285 then do
 say "ERROR: could not create semaphore: "||thesem_MODIFY '('aa
 exit
end
aa=eventsem_reset(thesem_MODIFY)
if aa<0 & aa<>-300 then do 
 say "ERROR: could not create modify semaphore: "||thesem_modify '('aa
 exit
end

/* Read unallowed ips */
badips.=''
badips.!wild=0
check_bad=0
if sreparams.!unallowed_File<>0 then do
   call read_unalloweds
   check_bad=badips.!do
   say '  # of unallowed IP entries: 'badips.!total '( 'badips.!wild ' wildcards)'
end 



/* set default servername and serveraddress */
serveraddr=sockgethostid()
rc=sockgethostbyaddr(serveraddr,'foo.!')
if rc=0 then
  serveraname=foo.!name
else
  servername=serveraddr

sz=stream(sreparams.!auditfile,'c','query size')
if sz>0 then do
  aa=charin(sreparams.!auditfile,sz,1)
  if aa='1a'x then sz=sz-1
end
else do
   sz=1
end 
foo=charout(sreparams.!auditfile,'0D0A'X||'0d0a'x||'**** Sre2000 started at '||time('n')||' '||date('n')||'  ',sz)  
foo=stream(sreparams.!auditfile,'c','close')
if stream(sreparams.!auditfile,'c','query exists')='' then do
    say "ERROR: Unable to create audit file: " sreparams.!auditfile
    exit
end
call make_auditqueue
say "  Using audit file: "sreparams.!auditfile

call make_cachequeue            /* used for request cache, and for md5s cache */

aa=sre_shutdown(0)
aa=sre_pause(0)

if sreparams.!init_proc<>0 & sreparams.!init_proc<>'' then do
  aproc=maindir'\'||strip(sreparams.!init_proc,,'\')
  say '  Calling custom initialization procedure  ('aproc') ....'
  foo=macrodrop('SRE2_INITPROC')
  foo=macroadd('SRE2_INITPROC',aproc,'b')
  if foo=0 then do
     say "Error: unable to load "aproc
     foo=macrodrop('SRE2_INITPROC')
     exit
  end 
  foo=sre2_initproc(sreparams.!port)
  if foo<>''  then say "  Status from initialization procedure: "bold||foo||normal
  foo=macrodrop('SRE2_INITPROC')
end


/* start daemon manager */


if sreparams.!daemon_manager<>0 & sreparams.!daemon_manager<>'' then do
  afilt2=maindir'\'||strip(sreparams.!daemon_manager,,'\')
  say '  Starting SRE daemon manager ('afilt2') ....'

 dmn_tid=rexxthread('f',afilt2,pid,thesem,maindir,sreparams.!PORT,servername)

/* wait for daemon manager to do it's business */
   say "  Waiting on initializations by daemon manager ..."
   do forever   /* wait for daemon monitor to say "go ahead" */
      foo=eventsem_wait(thesem,2000)
      if foo=0 then leave
   end    
end
aa=sre_shutdown()
if aa<>'0' & aa<>'' then do
   say "ERROR in daemon monitor: "aa
   exit
end

aa=eventsem_reset(thesem)

say bold 'Initialization completed for 'normal||'"'||sreparams.!server_software||'"'



return 0

/****************/
/* read unallowed_file */
read_unalloweds:

badips.=''
badips.!wild=0
badips.!do=1
badips.!total=0
ifoo=stream(sreparams.!unallowed_file,'c','query size')
if ifoo='' | ifoo=0 then do
    say "ERROR: could not find "sreparams.!unallowed_File 
   exit
end 
badips.!datetime=stream(sreparams.!unallowed_file,'c','query datetime')
sfoo=charin(sreparams.!unallowed_file,1,ifoo)
foo=stream(sreparams.!unallowed_file,'c','close')
do until sfoo=''
   parse var sfoo anip '0d0a'x sfoo ; 
   if anip='' then iterate
   anip=strip(word(anip,1))
   if abbrev(anip,';')=1 then iterate
   if pos('*',anip)>0 then do
       parse var anip a1 '*' .
       a1=strip(strip(a1),,'.')||'.*'
       badips.a1=1
       badips.!wild=badips.!wild+1
       badips.!total=badips.!total+1
   end 
   else do
       badips.anip=1
       badips.!total=badips.!total+1
   end
end

return 0

/**************************/
/* update SRE2000_CFG vars */
update_sre_ini:procedure expose sreparams. SRE2000_CFG maindir maindrive ,
                 default_gmtoffset0 basicqueue filter fast_filter

crlf='0d0a'x

/* wait until no requests are active, or 1 minute */
iok=0
do mm=1 to 60
  if sre_localvars('!CLIENTS')<>0 then do
     call syssleep(1)
     iterate
   end
   iok=1
   leave
end
if iok=0 then do 
   call call write_error "Server is busy -- unable to update parameters "
   return 0
end 


srevars_resetable='DATADIR LIMITBODY LIMITCLIENTS LIMITHEADER LIMITTIMETOTAL ' ,
        'LIMITTIMEWAIT LIMITTIMEINACTIVE ' ,
        'SERVER_SOFTWARE SERVER_SOFTWARE_SHORT REQ_CACHE_ENABLE REQ_CACHE_RECORD ' 

/* read initialization parameters */

foo=strip(translate(stream(SRE2000_CFG,'c','open read')))
if abbrev(foo,'READY')=0 then do
   call write_error 'INI error: Can not open initialization file'
   return 1
end 
isz=stream(SRE2000_CFG,'c','query size')
if (isz=0 | isz='') then do
   call write_error 'INI error: Empty or missing initialization file'
   return 1
end 

stuff=charin(SRE2000_CFG,1,isz)
foo=stream(SRE2000_CFG,'c','close')

/* now parse the file, eventually to save to SRE_SYSVARS cache */

do until stuff=''
   parse var stuff aline (crlf) stuff
   if aline='' then iterate
   aline=strip(translate(aline,' ','0d0a09'x))
   if abbrev(aline,';')=1 then iterate
   parse var aline avar '=' aval ; 
   avar=translate(strip(avar)) ; aval=strip(aval)
   if wordpos(avar,srevars_resetable)=0 then iterate  /* only change "resetable" variables */
   cname='!'||avar
   sreparams.cname=aval
end 

do mm=1 to words(srevars_resetable)
   aw=strip(word(srevars_resetable,mm))
   aw2='!'||aw
   foo=sre_sysvars(aw,sreparams.aw2)
   if abbrev(translate(foo),'ERROR')=1 then do
        call write_error 'Fatal INI error: unable to save 'aw' ('foo
        return 0
   end 
end 

/* determine gmtoffset using TZ? */
sreparams.!gmtoffset=fig_gmtoffset(tz,default_gmtoffset0)

call write_message "  Using GMT Offset = "sreparams.!gmtoffset
oo=sreparams.!GMTOFFSET
foo=sre_sysvars('GMTOFFSET',oo)

/* check existence of datadir */
datadir=sreparams.!DATADIR
datadir=translate(datadir,'\','/')
datadir=strip(strip(datadir,'t','\'))
if pos(':',datadir)=0 then do
   if abbrev(datadir,'\')=1 then do
        datadir=maindrive||datadir
   end
   else do
      datadir=maindir||'\'||datadir
   end /* do */
end
dd=dosisdir(datadir)
if dd=0 then do
   call write_error 'FATAL INI error: no such datadir ('datadir
   return 0
end

sreparams.!datadir=datadir
foo=sre_sysvars('DATADIR',datadir)


/* check limitbody and limitheader */
if datatype(sreparams.!limitbody)<>'NUM' then do
   call write_error 'FATAL INI error: LIMITBODY must be an integer > 0 '
   return 0
end 
sreparams.!limitbody=sreparams.!limitbody*1000

/* check limitbody and limitheader */
if datatype(sreparams.!limitheader)<>'NUM' then do
   call write_error 'FATAL INI error: LIMITHEADER must be an integer > 0 '
   exit
end 
sreparams.!limitheader=sreparams.!limitheader*1000

/* check cache_enable */
if sreparams.!req_cache_size=0 then sreparams.!req_cache_enable=0
if datatype(sreparams.!req_cache_record)<>'NUM' then sreparams.!req_cache_record=0


foo=macrodrop('SRE_FILTER')
foo=macroadd('SRE_FILTER',filter,'B')
if foo<>1 then do
    call sre_say 'ERROR: unable to re-load main filter ('filter
    leave
end

/* reload fast-filter? */
if fast_filter<>0 then do
      foo=macrodrop('SRE_FAST_FILTER')
      foo=macroadd('SRE_FAST_FILTER',fast_filter,'B')
      if foo<>1 then do
         call sre_say 'ERROR: unable to re-load fast filter ('fast_filter
         leave
      end
end


call write_reset ' (parameters updated at '||time('n')||')'

return 1

/******/
/* write to error line on status screen */
write_error:
 parse arg dog
 if length(dog)<68 then dog=left(dog,68,' ')
 call charout, '1b5b'x||'11;2H'||'1b'x||'[0;m'||dog
 call charout, '1b5b'x||'17;1H : '

 foo=sre_audit('SRE2000',strip(dog))
return 0

/******/
/* write to  stats line on status screen */
write_stats:
 parse arg dog
 if length(dog)<68 then dog=left(dog,68,' ')
 call charout, '1b5b'x||'7;2H'||'1b'x||'[0;m'||dog
 call charout, '1b5b'x||'17;1H : '
return 0

/******/
/* write to  message line on status screen */
write_message:
 parse arg dog
 dog=left(dog,68,' ')
 call charout, '1b5b'x||'13;2H'||'1b'x||'[0;m'||dog
 call charout, '1b5b'x||'17;1H : '

return 0



/******/
/* write to  say line on status screen */
write_reset:
 parse arg dog
 if length(dog)<35 then dog=left(dog,35,' ')
 call charout, '1b5b'x||'2;35H'||'1b'x||'[0;m'||dog
 call charout, '1b5b'x||'17;1H : '

return 0



/************/
make_auditqueue:procedure expose enmadd auditqueue

signal off syntax ; signal off error
signal on syntax name badq ; signal on error name badq

auditqueue=enmadd||'SRE_AUDIT'
a=rxqueue('s',auditqueue)
noq=1
queue 0
noq=0
badq:

signal on syntax name abend
signal on error name abend

if noq=1 then do
  foo=rxqueue('c',auditqueue)
  if upper(foo)<>upper(auditqueue) & sreparams.!verbose>0 then  do
        call write_error '  WARNING: queue error '||foo||auditqueue
  end
  a=rxqueue('s',auditqueue)
end
else do
  aa2=rxqueue('d',auditqueue)
  foo=rxqueue('c',auditqueue)
end

foo=sre_sysvars('AUDIT_QUEUE',auditqueue)

/* now make auditqueue semaphore */
auditsem='\SEM32\'||auditqueue
aa=eventsem_close(auditsem)
aa=eventsem_create(auditsem)
if aa<>0 & aa<>285 then do
 say "ERROR: could not create semaphore: "||auditsem '('aa
 exit
end
aa=eventsem_reset(auditsem)
if aa<0 & aa<>-300 then do 
 say "ERROR: could not create reset: "||auditsem '('aa
 exit
end


return 1


/************/
make_cachequeue:procedure expose enmadd cachequeue

signal off syntax ; signal off error
signal on syntax name badqc ; signal on error name badqc

cachequeue=enmadd||'SRE_CACHE'
a=rxqueue('s',cachequeue)
noq=1
queue 0
noq=0
badqc:

signal on syntax name abend
signal on error name abend

if noq=1 then do
  foo=rxqueue('c',cachequeue)
  if upper(foo)<>upper(cachequeue) & sreparams.!verbose>0 then  do
        call write_error '  WARNING: queue error '||foo||cachequeue
  end
  a=rxqueue('s',cachequeue)
end
else do
  aa2=rxqueue('d',cachequeue)
  foo=rxqueue('c',cachequeue)
end

foo=sre_sysvars('REQ_CACHE_QUEUE',cachequeue)
/* call pmprintf(" xxxxx req_cache_queue "foo) */

return 1






 /* ------------------------------------------------------------------ */
 /* function: Check if ANSI is activated                               */
 /*                                                                    */
 /* call:     CheckAnsi                                                */
 /*                                                                    */
 /* where:    -                                                        */
 /*                                                                    */
 /* returns:  1 - ANSI support detected                                */
 /*           0 - no ANSI support available                            */
 /*          -1 - error detecting ansi                                 */
 /*                                                                    */
 /* note:     Tested with the German and the US version of OS/2 3.0    */
 /*                                                                    */
 /*                                                                    */
 CheckAnsi: PROCEDURE
   thisRC = -1

   trace off
                         /* install a local error handler              */
   SIGNAL ON ERROR Name InitAnsiEnd

   "@ANSI 2>NUL | rxqueue 2>NUL"

   thisRC = 0

   do while queued() <> 0
     queueLine = lineIN( "QUEUE:" )
     if pos( " on.", queueLine ) <> 0 | ,                       /* USA */
        pos( " (ON).", queueLine ) <> 0 then                    /* GER */
       thisRC = 1
   end /* do while queued() <> 0 */

 InitAnsiEnd:
 signal off error
 RETURN thisRC


/***********/
/* determine gmtoffset given tz variable */
fig_gmtoffset:procedure expose basicqueue
parse upper arg atz,defg
atz=strip(atz)

signal on syntax name foo2 
signal on error  name  foo2
signal on failure name foo2
signal on notready name foo2

curqueue=rxqueue('S',basicqueue)
'@TZSET | RXQUEUE 2> nul '
do until queued()=0
   pull foo1
end 
foo=rxqueue('S',curqueue)

signal on syntax name abend
signal on error name abend
signal on failure name abend

foo1=space(foo1)
if abbrev(foo1,'THE CURRENT TIME ZONE')=0 then do
   foo=sre_audit('SRE2000',"ERROR Unexpected output from TZSET.EXE: "foo1)
   foo=sre_audit('SRE2000',"Using default gmtoffset= " defg,'111')
   return defg
end 
oo=word(foo1,words(foo1))
isneg=1
if left(oo,1)='-' then do 
  isneg=-1
  oo=substr(oo,2)
end 
if left(oo,1)='+' then do
   isneg=1
   oo=substr(oo,2)
end 
ii=length(oo)
minn=substr(oo,ii-1)
hrr=left(oo,ii-2)
aoff=-isneg*((hrr*3600)+(minn*60))
return aoff

foo2:
foo=sre_audit('Sre2000',"ERROR Unexpected output from TZSET.EXE: "foo1)
foo=sre_audit('Sre2000',"Using default gmtoffset= " defg,'111')
return defg

/*******/
showopts:
say
say "             "cy_ye"Sre2000 : An http/1.1 server for OS/2 "normal
say 
say cy_ye"Options: "normal
say bold"   no options:"normal" start the server, using options in "SRE2000_CFG
say bold"          ?  :"normal"  View this help screen "
say bold"  -i filename:"normal" Specify an initializaton file (instead of "SRE2000_CFG")"
say bold"  -m [filename] :"normal" Modify most important initialization parameters -- "
say     "                  do NOT start the server"
say "                  (if filename not specified, modify "SRE2000_CFG") "
say bold"  -mall [filename] :"normal" Modify all initialization parameters -- "
say "                     do NOT start the server"
say "                     (if filename not specified, modify "SRE2000_CFG") "

say bold"         -cvt  :"normal"  Convert a file from GoServe-syntax to Sre2000-syntax "

say 
say "Examples (assuming the Sre2000 working directory is D:\Sre2000)"
say   bold"   D:\Sre2000>Sre2000 "normal
say   bold"   D:\Sre2000>Sre2000 -i Sre2000.in2 "normal
say   bold"   D:\Sre2000>Sre2000  -cvt  "normal
say   bold"   D:\Sre2000>Sre2000  -m  "normal
say   bold"   D:\Sre2000>Sre2000  -mall  "normal

say
return 0
